# -*- coding: utf-8 -*-
'''
Copyright:	Schleifenbauer - 2018
Version:	1.1.0
Authors:	Laurent - laurent.schuermans@schleifenbauer.eu
			Schleifenbauer - support@schleifenbauer.eu

Permission to use, copy, modify, and/or distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright notice
and this permission notice appear in all copies.

This software is provided "as is" and Schleifenbauer disclaims all warranties
with regard to this software including all implied warranties of merchantability
and fitness. In no event shall Schleifenbauer be liable for any special, direct,
indirect, or consequential damages or any damages whatsoever resulting from loss
of use, data or profits, whether in an action of contract, negligence or other
tortious action, arising out of or in connection with the use or performance of
this software.
'''

# System imports
import queue
import threading
import socket
import urllib
import re
import ipaddress

# Library imports
from . library import requests

# Local imports
from . Helper import Helper as HLAPIHelper
from . DeviceManager import DeviceManager

class NetworkScanner():
	def __init__(self, hlapi_instance, input, http_port, webapi_user):
		self.hlapi = hlapi_instance
		self.input = input
		self.mode = None
		self.threads = self.hlapi.getConfig('max_threads')

		# Connection variabled
		self.http_port = http_port
		self.webapi_user = webapi_user

		self.q = queue.Queue()
		self.result = []
		self.inputAccepted = True
		self.parseInput(input)
		self.partial_ip = self.parseSubnet(input)
		self.progress = HLAPIHelper.createProgressManager(self.hlapi, name='NetworkScanner')

	def parseInput(self, input):
		# Determine IP mode
		ipv4_address = re.compile('^(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')
		ipv6_address = re.compile('^(?:(?:[0-9A-Fa-f]{1,4}:){6}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|::(?:[0-9A-Fa-f]{1,4}:){5}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){4}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){3}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,2}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){2}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,3}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}:(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,4}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,5}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}|(?:(?:[0-9A-Fa-f]{1,4}:){,6}[0-9A-Fa-f]{1,4})?::)$')

		temp = input.replace('*', '0')
		if len(re.findall(ipv4_address, temp)) == 1:
			self.mode = 'ipv4'
			if self.hlapi.debug: print("Received IPv4 address")
		elif len(re.findall(ipv6_address, temp)) == 1:
			self.mode = 'ipv6'
			if self.hlapi.debug: print("Received IPv6 address")
		else:
			if self.hlapi.debug: print("Invalid IP address")

	# Remove '*' from subnet input
	def parseSubnet(self, ip):
		if self.mode == 'ipv4':
			return ip.replace('*', '')
		elif self.mode == 'ipv6':
			return ipaddress.ip_address(ip.replace('*', '')).exploded
		else:
			self.inputAccepted = False
			return None

	def startScan(self):
		if self.hlapi.debug: print("Starting network scan for", str(self.partial_ip)+"*")
		self.progress.reset()
		t = threading.Thread(target=self.scan, args=[])
		self.progress.addThreadWatch(t)
		t.start()

	# Add all target IP addri to a queue and let x threads evaluate them.
	def scan(self):
		if not self.inputAccepted:
			self.result = None
			self.progress.abort()
			return

		if "*" in self.input:
			self.result = []
			# Add IPv4 range to queue
			if self.mode == 'ipv4':
				self.progress.setTarget(254)
				for target in range(1, 255):
					self.q.put(self.partial_ip+str(target))

			# Add IPv6 range to queue, currently only the last group can be scanned
			elif self.mode == 'ipv6':
				pass
				# start = 0x0000
				# end = 0xFFFF
				# self.progress.setTarget(int(0xFFFF))
				# for target in range(start, end+1):
				#	append = hex(target).rstrip("L").lstrip("0x") or "0"
				#	self.q.put(self.partial_ip+str(append))


			# Start worker threads
			if self.hlapi.debug: print("Starting", self.threads, "threads for scanning of", self.mode, "subnet")
			for x in range(self.threads):
				t = threading.Thread(target=self.scanThread)
				self.progress.addThreadWatch(t)
				t.start()

		else:
			self.progress.setTarget(1)
			self.pollHost(self.input)
			self.progress.addProgress(1)

	# Aborts scanning, clear queue
	def abortScan(self):
		self.q.mutex.acquire()
		self.q.queue.clear()
		self.q.all_tasks_done.notify_all()
		self.q.unfinished_tasks = 0
		self.q.mutex.release()
		self.progress.abort()

	def getResult(self):
		return self.result

	# Adds an IP address to self.result if a connection to self.port or http can be established.
	def pollHost(self, target):
		if self.mode == 'ipv4':
			af = socket.AF_INET
		elif self.mode == 'ipv6':
			af = socket.AF_INET6

		if not self.progress.isAborted() or self.progress.isError():
			s = socket.socket(af, socket.SOCK_STREAM)
			try:
				# First try IPAPI
				info = socket.getaddrinfo(target, self.http_port, af, socket.SOCK_STREAM)
				s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
				s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
				s.settimeout(self.hlapi.getConfig('ipapi_timeout'))
				s.connect( info[0][4] )
				s.settimeout(None)
				if self.hlapi.debug: print(target, "IPAPI connection established")
				self.result.append(target)
			except Exception as e:
				# Try WEBAPI on fail
				headers = {'Content-type': 'application/x-www-form-urlencoded'}
				baseurl = 'http://'+str(target)
				for suffix in ['/userid', '/session/userid']:
					try:
						stitched = b''
						response = requests.post(baseurl+suffix, timeout=self.hlapi.getConfig('webapi_timeout'), headers=headers, stream=True, data={'user': self.webapi_user})
						httpcode = response.status_code
						if httpcode == 200:
							for chunk in response.iter_content(chunk_size=None):
								stitched += chunk
							data = urllib.parse.parse_qs(stitched.decode("utf-8"), True)
							if 'userid' in data.keys():
								if self.hlapi.debug: print(target, "WEBAPI connection established")
								self.result.append(target)
					except (requests.exceptions.RequestException, socket.error, socket.timeout) as e:
						pass
			s.close()

	# Worker thread, takes an IP from the queue and evaluates it.
	def scanThread(self):
		while self.progress.isRunning():
			try:
				current_target = self.q.get_nowait()
				while self.progress.isRunning():
					try:
						self.pollHost(current_target)
						break
					except:
						if self.hlapi.debug: print("Too many open sockets, retry in one second...")
						time.sleep(1)
				if not self.progress.isAborted():
					self.q.task_done()
					self.progress.addProgress(1)
			except queue.Empty:
				pass
