# -*- coding: utf-8 -*-
'''
Copyright:	Schleifenbauer - 2018
Version:	1.1.0
Authors:	Laurent - laurent.schuermans@schleifenbauer.eu
			Schleifenbauer - support@schleifenbauer.eu

Permission to use, copy, modify, and/or distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright notice
and this permission notice appear in all copies.

This software is provided "as is" and Schleifenbauer disclaims all warranties
with regard to this software including all implied warranties of merchantability
and fitness. In no event shall Schleifenbauer be liable for any special, direct,
indirect, or consequential damages or any damages whatsoever resulting from loss
of use, data or profits, whether in an action of contract, negligence or other
tortious action, arising out of or in connection with the use or performance of
this software.
'''

# System imports
import sys
import getopt
import importlib
import ast
import time
import threading
import json

# Local imports
from . Helper import Helper as HLAPIHelper
from . DeviceManager import DeviceManager
from . managers.MultiReadWrite import MultiReadWrite
from . NetworkScanner import NetworkScanner

class HLAPI():
	"""
	This class handles HLAPI configuration and command line argument parsing if needed.
	python3.5 -m hlapi -t 192.168.1.1;192.168.1.2#223 -a read -d idfwvs,idspdt
	"""

	DEFAULT_CONFIG = {
		'ipapi_timeout':	5,
		'ipapi_yield':		0.1,
		'ipapi_port':		7783,
		'webapi_timeout':	5,
		'scan_timeout':		10,
		'cache_expire':		1000,
		'file_compat_nr':	4,
		'downshift_tries':	5,
		'max_threads':		52,
	}

	DEFAULT_PROTOCOL_ORDER = ['WEBAPI', 'IPAPI']

	def __init__(self, debug=False):
		self.config = None
		self.debug = debug
		self.setConfig(HLAPI.DEFAULT_CONFIG)

	def getConfig(self, record=None):
		if record is not None:
			val = self.config.get(record, None)
			if val is None:
				self.config[record] = HLAPI.DEFAULT_CONFIG.get(record, None)
				return self.config[record]
			return val
		return self.config

	def setConfig(self, data):
		self.config = data

# Prints percentage until task is done
def showPercentage(progressManager, silent):
	progressManager.waitForInit()
	last_percentage = 0
	if not silent: print("0%")
	while progressManager.isRunning():
		percentage = progressManager.getStatus()[1]
		if percentage > last_percentage+10:
			if not silent: print(str(percentage)+"%")
			last_percentage = percentage
		time.sleep(0.1)
	if not silent: print("100%")
	progressManager.closeThreads()

# Convert Python dictionary to JSON
def returnResult(result, pretty):
	if pretty:
		print(json.dumps(result, indent=4))
	else:
		print(json.dumps(result))
	sys.exit(0)

def parseReadInput(input):
	result = []
	splitted = input.split(',')
	for mnemonic in splitted:
		result.append(mnemonic)
	return result

def commandLine(argv):
	debug = False
	silent = False
	pretty = False
	read = None
	targetPath = None

	# Parse command line arguments
	# -t target JSON file path (required)
	# -r "read", comma separated list of register mnemonics (required)
	# -s (silent mode, only outputs result)
	# --debug (enables HLAPI debug mode)
	# --pretty (pretty-prints json output to console)
	try:
		opts, args = getopt.getopt(sys.argv[1:], "t:r:s", ['debug', 'pretty'])
	except getopt.GetoptError:
		print('Invalid arguments')
		sys.exit(0)

	for opt, arg in opts:
		if opt == '-t':
			targetPath = arg
		elif opt == '-r':
			read = arg.lower()
		elif opt == '-s':
			silent = True
		elif opt == '--debug':
			debug = True
		elif opt == '--pretty':
			pretty = True

	if debug: silent = False
	targets = None

	# Load targets from path
	try:
		data = json.load(open(targetPath))
		targets = HLAPIHelper.organizeInterfaces(data)
		if not silent: print("Filtered out", len(data.keys())-len(targets.keys()), "interfaces (duplicate entries)")
	except:
		print('Input JSON error')
		sys.exit(0)

	# Check preset input
	if read is None:
		print('No registers given')
		sys.exit(0)

	registers = parseReadInput(read)
	if registers is None:
		print('Could not parse register input')
		sys.exit(0)

	# Convert SUBNETS in input to IPs
	hlapi = HLAPI(debug=debug)
	scannedIPs = []
	# Targets is now a dict containing SUBNETS, IPs and DIRECT (IP#UNIT)
	subnets = {k:v for (k, v) in targets.items() if '*' in k}

	# Scan each subnet
	for interface, parameters in subnets.items():
		if not silent: print("Scanning (network)", interface)

		ns = NetworkScanner(hlapi, interface, parameters['webapi_port'], parameters['webapi_user'])
		if ns.inputAccepted:
			ns.startScan()
		else:
			break

		showPercentage(ns.progress, silent)

		if ns.progress.isAborted() or ns.progress.isError():
			result = {
				'status': 'fail',
				'message': 'An error occured while network scanning.',
				'unknown_devices': None,
				'renumber_devices': None
			}
			returnResult(result, pretty)

		elif ns.progress.isDone():
			data = ns.getResult()
			if not silent: print(interface, "->", data)
			if data is not None:
				scannedIPs += data

	# Remove SUBNETS from targets (replace by scanned IPs with same credentials)
	targets = HLAPIHelper.subnetsToIPs(targets, scannedIPs)

	# Start device identifcation
	deviceManager = DeviceManager(hlapi)
	deviceManager.startLoadInterfaces(targets)

	error = False
	unknownDevices = None
	renumberDevices = None

	if not silent: print("Identifying / scanning (databus)", len(targets.keys()), "targets")
	p = deviceManager.progress
	showPercentage(p, silent)

	if p.isAborted() or p.isError():
		result = {
			'status': 'fail',
			'message': 'An error occured during device identification.',
			'unknown_devices': None,
			'renumber_devices': None
		}
		returnResult(result, pretty)

	if p.isDone():
		# List of targets (IPs and/or direct devices is now converted to a list of Device objects)
		devices = deviceManager.getResult('raw')
		renumberDevices = deviceManager.result[3]
		unknownDevices = deviceManager.unknownDevices
		if not silent: print("Found", len(devices), "devices")

		if len(renumberDevices) > 0:
			# Fail when one or more devices need to be renumbered
			result = {
				'status': 'fail',
				'message': 'Some devices could not be identified or need to be renumbered.',
				'unknown_devices': ["#".join([str(x), str(y)]) for x,y in unknownDevices],
				'renumber_devices': ["#".join([str(x), str(y)]) for x,y in renumberDevices]
			}
			returnResult(result, pretty)

		else:
			# Read the required registers from all devices
			if not silent: print("Reading registers...")
			multiReadWrite = MultiReadWrite(hlapi, devices)
			multiReadWrite.readAll(registers, block=False)
			p = multiReadWrite.progress
			showPercentage(p, silent)

			if p.isAborted() or p.isError():
				result = {
					'status': 'fail',
					'message': 'An error occured in MultiReadWrite.readAll().',
					'unknown_devices': None,
					'renumber_devices': None
				}
				returnResult(result, pretty)

			if p.isDone():
				result = {
					'status': 'ok',
					'data': HLAPIHelper.makeOrdered(multiReadWrite.result),
					'unknown_devices': ["#".join([str(x), str(y)]) for x,y in unknownDevices],
					'renumber_devices': ["#".join([str(x), str(y)]) for x,y in renumberDevices]
				}
				returnResult(result, pretty)

if __name__ == "__main__":
	commandLine(sys.argv[1:])
