# -*- coding: utf-8 -*-
'''
Copyright:	Schleifenbauer - 2018
Version:	1.1.0
Authors:	Laurent - laurent.schuermans@schleifenbauer.eu
			Schleifenbauer - support@schleifenbauer.eu

Permission to use, copy, modify, and/or distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright notice
and this permission notice appear in all copies.

This software is provided "as is" and Schleifenbauer disclaims all warranties
with regard to this software including all implied warranties of merchantability
and fitness. In no event shall Schleifenbauer be liable for any special, direct,
indirect, or consequential damages or any damages whatsoever resulting from loss
of use, data or profits, whether in an action of contract, negligence or other
tortious action, arising out of or in connection with the use or performance of
this software.
'''

# System imports
import os
import queue
import struct
import threading
import time
import re

# Local imports
from .. devices.AbstractDevice import AbstractDevice
from .. devices.Devices import *
from . ResetManager import ResetManager
from . MultiReadWrite import MultiReadWrite
from .. Helper import Helper as HLAPIHelper
from .. communication.Communicator import Communicator
from .. communication.IPAPIFramer import *
from .. communication.IPAPIProtocol import IPAPIProtocol

class FirmwareManager(MultiReadWrite):
	def __init__(self, hlapi_instance, devices):
		super(FirmwareManager, self).__init__(hlapi_instance, devices)
		self.hlapi = hlapi_instance
		self.progress = {}
		self.abort = False
		self.tasks = 0
		self.tasksDone = 0
		self.q = queue.Queue()
		self.threads = []

	# binpath can be a single .bin path for multiple devices or a dictionary of bin paths per device hardware id
	def startFirmwareUpgrade(self, data, deepupdate=False):
		# data = dict {'ip': [(version, file1), (version, file2)] (auto broadcast), 'ip:unit': [(version, file1)], ...}
		self.tasks = 0
		self.tasksDone = 0
		self.progress = {}
		self.abort = False

		threads = 20
		bundledDevices = HLAPIHelper.bundleDevicesByCommunicator(self.devices)

		if bundledDevices is not None:
			for x in range(threads):
				t = threading.Thread(target=self.firmwareUpgrade, args=[x, self.q, data, deepupdate])
				self.threads.append(t)
				t.start()

			for bundle in bundledDevices:
				self.q.put( bundle )

		for uid, d in data.items():
			self.tasks += len(d[1])
			for target in d[1]:
				version = target[0]
				self.setProgress(uid, version, 'Queued', 0)

		if self.hlapi.debug: print("Num tasks:", self.tasks)

	def abortAll(self):
		if self.hlapi.debug: print("Aborting firmware upgrade")
		self.abort = True
		self.q.mutex.acquire()
		self.q.queue.clear()
		self.q.all_tasks_done.notify_all()
		self.q.unfinished_tasks = 0
		self.q.mutex.release()
		for t in self.threads:
			t.join()

	def setProgress(self, uid, version, status, percentage):
		if uid in self.progress:
			current = self.progress[uid]
			findExistingEntry = None
			i = 0
			for x in current:
				if x[1] == version:
					findExistingEntry = i
					break
				i += 1
			if findExistingEntry is not None:
				self.progress[uid][findExistingEntry] = [status, version, percentage]
			else:
				self.progress[uid].append([status, version, percentage])
		else:
			self.progress[uid] = [[status, version, percentage]]

	def isDone(self):
		if (self.tasks > 0 and self.tasks == self.tasksDone) or self.abort:
			if self.hlapi.debug: print("Firmware updater done, abort:", self.abort)
			return True
		return False

	def firmwareUpgrade(self, tid, q, data, deepupdate):
		regex = '^(?P<oem>SP|MK)FW-(?P<version>\d{4})-(?P<checksum>[0-9A-F]{8})-(?P<crc>[0-9A-F]{4}).(?P<extra>.*)\.bin$'
		while not self.abort:
			try:
				bundle = q.get_nowait()
			except queue.Empty:
				time.sleep(1)
				continue
			if self.hlapi.debug: print("Thread", tid, "received bundle:", bundle)
			direct = []
			broadcast = []
			for device in bundle:
				if device.uid in data:
					dData = data.get(device.uid, None)
					if dData is not None:
						direct.append((dData[0], dData[1]))
				elif device.ip in data:
					dData = data.get(device.ip)
					if dData not in broadcast:
						broadcast.append((dData[0], dData[1]))
			if self.hlapi.debug: print(tid, "DIRECT:", direct)
			if self.hlapi.debug: print(tid, "BROADCAST:", broadcast)

			allDevices = direct + broadcast

			# All communicators in bundle are the same
			ipapi = allDevices[0][0].communicator.getProtocol('IPAPI')
			if ipapi is not None:
				for device in allDevices:
					if self.abort:
						break
					deviceData = {}
					deviceData['device'] = device[0]
					deviceData['broadcast'] = device in broadcast
					deviceData['paths'] = device[1]

					for target in deviceData['paths']:
						version = target[0]
						path = target[1]
						if self.abort:
							break
						if deviceData['broadcast']:
							currentID = deviceData['device'].ip
						else:
							currentID = deviceData['device'].uid

						self.setProgress(currentID, version, 'Preparing', 0)
						if self.hlapi.debug: print(str(tid)+":", "Upgrading", currentID, "(BR="+str(deviceData['broadcast'])+")", "to", str(version))
						filename = os.path.basename(path)
						match = re.match(regex, filename, re.IGNORECASE)
						if not match:
							if self.hlapi.debug: print(tid, 'Invalid filename')
							break

						groups = match.groupdict()
						file_version = int(groups['version'], base=10)
						file_checksum = int(groups['checksum'], base=16)
						file_crc = int(groups['crc'], base=16)

						# Split binary file into blocks of size 256
						deviceData['blocks'] = []
						deviceData['bytesize'] = 0
						with open(path, mode='rb') as binf:
							while True:
								block = binf.read(256)
								if block: # will be empty string (thus False) if EOF reached
									deviceData['blocks'].append(block)
									deviceData['bytesize'] += len(block)
								else:
									break
						deviceData['num_blocks'] = len(deviceData['blocks'])

						# Write firmware info to 'upvers' register (10000)
						# [version][checksum][crc][numberOfBlocks][size]
						#
						# < = little-endian
						# H = unsigned short integer size 2
						# L = unsigned long integer size 4
						fw_info = struct.pack('<HLHHL', file_version, file_checksum, file_crc, deviceData['num_blocks'], deviceData['bytesize'])

						# Temporarily increase IPAPI timeout
						backupIPAPITimeout = ipapi.timeout
						ipapi.timeout = 50 # TODO: why 50? can it take 50 seconds to re-acquire connection?
						if self.hlapi.debug: print(tid, 'Increased IPAPI timeout')

						# Delay can be reduced to about 0.1 for non-ringmaster devices that are directly addressable (ethernet, no databus)
						# however, the hPDU mode cannot be read through the IPAPI making it impossible to determine whether the device
						# is behind a gateway or another hPDU...
						delay = 0.3

						# Since each thread handles a bundle of devices using the same communicator,
						# we can only fast-update if the bundle contains a single device (no databus)
						if deviceData['device'].isHybridMaster():
							delay = 0.1

						if self.hlapi.debug: print(tid, "Packet delay set to", delay)

						ipapi._openSocket()
						if ipapi.conn:
							f = Frame()
							if deviceData['broadcast']:
								# Reboot devices before firmware updating to prevent PDU getting stuck at 'RETRY'.
								rsManager = ResetManager(self.hlapi, deviceData['device'])
								rsManager.rebootDevice()

								# Send 3 dummy bytes so a gateway gives us the full socket without yielding.
								for x in range(3, 0, -1):
									if self.hlapi.debug: print(tid, "Dummy broadcast "+str(x))
									d = Frame()
									d.Fields = BroadcastWriteRegisterFields
									d.setField(COMMAND, d.CMD_BRWRITE[0])
									d.setField(REGISTERSTART, 10000)
									d.setField(REGISTERLENGTH, 14)
									d.setField(DATA, struct.pack('<HLHHL', 0, file_checksum, file_crc, x, 3))
									ipapi.writeRaw(d)
									time.sleep(delay)

								f.Fields = BroadcastWriteRegisterFields
								f.setField(COMMAND, f.CMD_BRWRITE[0])
							else:
								f.Fields = WriteRegisterFields
								f.setField(COMMAND, f.CMD_WRITE[0])
								f.setField(UNITADDRESS, deviceData['device'].unitAddress)

							f.setField(REGISTERSTART, 10000)
							f.setField(REGISTERLENGTH, 14)
							f.setField(DATA, fw_info)

							if deepupdate and deviceData['broadcast']:
								deviceData['writeRepeats'] = 3
							else:
								deviceData['writeRepeats'] = 1

							self.setProgress(currentID, version, 'Writing header', 0)
							if self.hlapi.debug: print(tid, "Writing header:", file_version, file_checksum, file_crc, deviceData['num_blocks'], deviceData['bytesize'])

							for x in range(3):
								writer = ipapi.writeRaw(f)
								time.sleep(1)

							self.setProgress(currentID, version, 'Erasing flash', 0)

							# PDU Erase flash
							time.sleep(14)

							# Now that we have aquired a lock on the end device socket, temporarily disable API yielding.
							backupIPAPIYieldSeconds = ipapi.yieldSeconds
							ipapi.yieldSeconds = 0
							if self.hlapi.debug: print(tid, 'Disabled IPAPI yielding')

							# Write blocks to 'upblnr' (10100) register
							# [dataBlockNumber][dataBlock]
							# Each block write requires an updated block number (size 2) and data (size 256)
							if self.hlapi.debug: print(tid, 'Writing blocks...')
							if self.hlapi.debug: print(tid, 'ETA:', int(deviceData['num_blocks'])*deviceData['writeRepeats']*delay+30, 'seconds')
							# gen 1 broadcast ± minutes
							# gen 1 broadcast deep update ± minutes
							# gen 2/3 broadcast ±10 minutes
							# gen 2/3 deep update broadcast ±27 minutes
							# gen 2/3 direct ±3 minutes

							# The next three loops are in place to ensure packets being sent as 123123123 instead of 111222333
							for i in range(0, deviceData['num_blocks'], deviceData['writeRepeats']):
								if self.abort:
									break
								if not ipapi.conn:
									break
								for x in range(deviceData['writeRepeats']):
									if self.abort:
										break
									for y in range(deviceData['writeRepeats']):
										if self.abort:
											break

										currentBlock = i+y

										# The last block should only be written once, otherwise the pdu will reboot and all that.
										# Upon writing the last block, it will be repeated, see below.
										if i == deviceData['num_blocks']-1 and (x != 0 or y != 0):
											continue

										# Current block data
										block = deviceData['blocks'][currentBlock]
										# < = Little-endian
										# H = unsigned short integer size 2
										# s = char[] string (256 bytes)
										data = struct.pack('<H256s', currentBlock, block)

										# Same reason as dummy header writes, we need to re-aquire the lock on a gateway's socket.
										if i == 0 and x == 0: # first 3 writes
											time.sleep(1)

										# Prepare frame
										f = Frame()
										if deviceData['broadcast']:
											f.Fields = BroadcastWriteRegisterFields
											f.setField(COMMAND, f.CMD_BRWRITE[0])
										else:
											f.Fields = WriteRegisterFields
											f.setField(COMMAND, f.CMD_WRITE[0])
											f.setField(UNITADDRESS, deviceData['device'].unitAddress)

										f.setField(REGISTERSTART, 10100)
										f.setField(REGISTERLENGTH, 258)
										f.setField(DATA, data)

										time_start = time.time()

										# Send frame
										if (currentBlock+1) % 100 == 0:
											if self.hlapi.debug: print(tid, 'Writing block', currentBlock+1, "of", deviceData['num_blocks'])
											if not isinstance(deviceData['device'], Gateway):
												if self.hlapi.debug: print(tid, 'Checking connection...')
												ipapi._closeSocket()
												ipapi._openSocket()
												
										ipapi.writeRaw(f)

										# Rewrite last packet because classic PDUs behing a bridged device have no patience
										if currentBlock == deviceData['num_blocks']-1:
											# Quick fix for PDU queue deadlock bug
											time.sleep(6)
											for z in range(deviceData['writeRepeats']):
												ipapi.writeRaw(f)
												time.sleep(delay)

										noSleep = False

										if not deviceData['broadcast']:
											# Block until ACK received to prevent multiple packets on databus at the same time
											# The gateway will expect classic PDUs to return ACK but they will return a timeout
											# error code for each sent block.
											if not isinstance(deviceData['device'], cPDU):
												response = ipapi.readRaw()
												if response is None or response == 'TIMEOUT':
													if self.hlapi.debug: print("No response")
												else:
													noSleep = True

										# Prevent percentage from going down
										if x == 0 and y == 0:
											self.setProgress(currentID, version, 'Writing FW', int(currentBlock/(deviceData['num_blocks']-1)*100))

										sleepTime = 0
										# The first few blocks are sent with a longer delay
										if i == 0 and x == 0:
											sleepTime = 1

										# Calculate total time from block write to now (repeats, delays, response)
										# and sleep if time is less than expected
										time_elapsed = time.time()-time_start

										if not noSleep and time_elapsed < delay:
											sleep = max(sleepTime, delay-time_elapsed) # either 0, 1 or delay-time_elapsed
											time.sleep(sleep)
						else:
							self.setProgress(currentID, version, 'Error: No IPAPI', 0)

						if ipapi.conn and not self.abort:
							if self.hlapi.debug: print(tid, 'Verify...')
							self.setProgress(currentID, version, 'FW verification', 100)

							# PDU Verify firmware
							time.sleep(20)

							if self.hlapi.debug: print(tid, 'Rebooting...')
							self.setProgress(currentID, version, 'Rebooting', 100)

							# Reboot devices and reset communicator
							rsManager = ResetManager(self.hlapi, deviceData['device'])
							rsManager.rebootDevice()

							if not deviceData['broadcast']:
								# Check if device has flashed to new firmware
								newfw = deviceData['device'].read('idfwvs', 'single', cache=False)
								if newfw is not None:
									idfwvs = newfw.get('idfwvs', None)
									if idfwvs == version:
										if self.hlapi.debug: print(tid, 'Success')
										self.setProgress(currentID, version, 'Success', 100)
									else:
										if self.hlapi.debug: print(tid, 'Fail?')
										self.setProgress(currentID, version, 'Update or reboot failed', 100)
								else:
									self.setProgress(currentID, version, 'Unknown', 100)
							else:
								self.setProgress(currentID, version, 'Done', 100)

						time.sleep(3)
						self.tasksDone += 1

						# Restore API yielding
						ipapi.yieldSeconds = backupIPAPIYieldSeconds
						if self.hlapi.debug: print(tid, 'IPAPI yielding restored')
						# Restore API timeout
						ipapi.timeout = backupIPAPITimeout
						if self.hlapi.debug: print(tid, 'IPAPI timeout restored')
						# Flush
						ipapi._closeSocket()
			else:
				for device in allDevices:
					if device in broadcast:
						currentID = device[0].ip
					else:
						currentID = device[0].uid
					for target in device[1]:
						version = target[0]
						self.setProgress(currentID, version, 'Error: No IPAPI', 0)
					self.tasksDone += len(device[1])
			if not self.abort:
				q.task_done()
