# -*- coding: utf-8 -*-
'''
Copyright:	Schleifenbauer - 2018
Version:	1.1.0
Authors:	Laurent - laurent.schuermans@schleifenbauer.eu
			Schleifenbauer - support@schleifenbauer.eu

Permission to use, copy, modify, and/or distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright notice
and this permission notice appear in all copies.

This software is provided "as is" and Schleifenbauer disclaims all warranties
with regard to this software including all implied warranties of merchantability
and fitness. In no event shall Schleifenbauer be liable for any special, direct,
indirect, or consequential damages or any damages whatsoever resulting from loss
of use, data or profits, whether in an action of contract, negligence or other
tortious action, arising out of or in connection with the use or performance of
this software.
'''

# System imports
import threading
import queue

# Local imports
from .. Helper import Helper as HLAPIHelper
from .. spdm.RegisterHelper import RegisterHelper

# MultiReadWrite class, each high level manager inherits this class
#
class MultiReadWrite(object):

	# Set self.devices to a list of device instances
	def __init__(self, hlapi_instance, devices, progress=None):
		self.hlapi = hlapi_instance
		if isinstance(devices, list):
			self.devices = devices
		else:
			self.devices = [devices]

		self.result = None

		if progress is None:
			self.progress = HLAPIHelper.createProgressManager(self.hlapi, name='MultiReadWrite')
		else:
			self.progress = progress

	def readSingle(self, mnemonic):
		result = self.readAll([mnemonic])
		# We know each UID only has one value (single register)
		for key, value in result.items():
			result[key] = HLAPIHelper.extract(value)
		return result

	def writeSingle(self, mnemonic, data):
		dataMap = {}
		for device in self.devices:
			dataMap[device.getUid()] = {mnemonic: data}
		result = self.writeAll(dataMap)
		# We know each UID only has one value (single register)
		for key, value in result.items():
			result[key] = HLAPIHelper.extract(value)
		return result

	# Wrapper function for HLAPI consistency
	def startReadAll(self, mnemonics):
		self.readAll(mnemonics, block=False)

	def readAll(self, mnemonics, block=True):
		self.result = {}
		self.progress.reset()

		bundled = HLAPIHelper.bundleDevicesByCommunicator(self.devices)
		if self.hlapi.debug: print("Starting", len(bundled), "threads for mass read of", len(mnemonics), "registers on", len(self.devices), "devices.")
		self.progress.setTarget(len(mnemonics) * len(self.devices))

		worker_queue = queue.Queue()
		for bundle in bundled:
			worker_queue.put(bundle)

		max_threads = self.hlapi.getConfig('max_threads')
		for i in range(min(len(bundled), max_threads)):
			t = threading.Thread(target=self._readThread, args=[worker_queue, mnemonics])
			self.progress.addThreadWatch(t)
			t.start()

		if block:
			self.progress.closeThreads()
			return HLAPIHelper.makeOrdered(self.result)

	def _readThread(self, q, mnemonics):
		while self.progress.isRunning():
			try:
				devices = q.get_nowait()
			except queue.Empty:
				return

			for device in devices:
				self.result[device.getUid()] = {}
				for mnemonic in mnemonics:
					if self.progress.isAborted() or self.progress.isError():
						q.mutex.acquire()
						q.queue.clear()
						q.all_tasks_done.notify_all()
						q.unfinished_tasks = 0
						q.mutex.release()
						return
					data = device.read(mnemonic, 'single')
					if data is not None:
						self.result[device.getUid()][mnemonic] = data.get(mnemonic, None)
					else:
						self.result[device.getUid()][mnemonic] = None

					self.progress.addProgress(1)
			q.task_done()

	# Wrapper function for HLAPI consistency
	def startWriteAll(self, dataMap):
		self.writeAll(dataMap, block=False)

	# dataMap: {UID: {mnemonic:value, mnemonic:value}, UID: {mnemonic:value, mnemonic:value}, etc.}
	def writeAll(self, dataMap, block=True):
		self.result = {}
		self.progress.reset()

		bundled = HLAPIHelper.bundleDevicesByCommunicator(self.devices)
		if self.hlapi.debug: print("Starting", len(bundled), "threads for mass write on", len(self.devices), "devices.")

		num_writes = 0
		for uid, values in dataMap.items():
			num_writes += len(values.keys())
		self.progress.setTarget(num_writes)

		worker_queue = queue.Queue()
		for bundle in bundled:
			worker_queue.put(bundle)

		max_threads = self.hlapi.getConfig('max_threads')
		for i in range(min(len(bundled), max_threads)):
			t = threading.Thread(target=self._writeThread, args=[worker_queue, dataMap])
			self.progress.addThreadWatch(t)
			t.start()

		if block:
			self.progress.closeThreads()
			return HLAPIHelper.makeOrdered(self.result)

	# Same interface devices in a list
	def _writeThread(self, q, dataMap):
		while self.progress.isRunning():
			try:
				devices = q.get_nowait()
			except queue.Empty:
				return

			for device in devices:
				dataMapEntry = dataMap.get(device.getUid(), {})

				for mnemonic, value in dataMapEntry.items():
					if self.progress.isAborted() or self.progress.isError():
						q.mutex.acquire()
						q.queue.clear()
						q.all_tasks_done.notify_all()
						q.unfinished_tasks = 0
						q.mutex.release()
						return
					data = device.write(mnemonic, 'single', value)
					if data is not None:
						status = data
					else:
						status = False

					deviceUid = device.getUid()
					if not deviceUid in self.result:
						self.result[deviceUid] = {}

					self.result[deviceUid][mnemonic] = status
					self.progress.addProgress(1)
			q.task_done()
